###################################################################
#
# flict - FOSS License Compatibility Tool
#
# SPDX-FileCopyrightText: 2021 Jens Erdmann
# SPDX-FileCopyrightText: 2022 Henrik Sandklef
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
###################################################################


from flict.flictlib.return_codes import FlictError, ReturnCodes

from flict.flictlib.arbiter import Arbiter
from flict.flictlib.format.factory import FormatterFactory
from flict.flictlib.project.reader import ProjectReaderFactory

import json


class FlictImpl:

    def __init__(self, args) -> None:
        self._args = args
        self.arbiter = self._get_arbiter()
        self._formatter = FormatterFactory.formatter(args.output_format)

    def merge_license_db(self):
        return self.arbiter.extend_license_db(self._args.license_file)

    def display_compatibility(self):
        inter_compats = self.arbiter.check_compatibilities(self._args.license_expression)
        return self._formatter.format_compats(inter_compats)

    def simplify(self):
        simplified = self.arbiter.simplify_license(" ".join(self._args.license_expression))
        return self._formatter.format_simplified(simplified)

    def suggest_outbound_candidate(self):
        if self._args.extended_licenses:
            licenses = self.arbiter.supported_licenses()
        else:
            licenses = self.arbiter.licenses(" ".join(self._args.license_expression))
        outbounds = []
        try:
            for outbound in licenses:
                compats = self.arbiter.inbounds_outbound_check(outbound, self._args.license_expression)
                compatible = (compats['compatibility'] == "Yes")
                if compatible:
                    outbounds.append(outbound)

            outbounds.sort()
            return self._formatter.format_outbound_license(outbounds)

        except BaseException:
            raise FlictError(ReturnCodes.RET_INVALID_EXPRESSSION,
                             f'Invalid license expression: {self._args.license_expression}')

    def list_licenses(self):
        licenses = list(self.arbiter.supported_licenses())
        licenses.sort(key=str.lower)
        return self._formatter.format_support_licenses(licenses)

    def _handle_lico_project(self, reader, project_file, formatter):
        project = reader.read_project(project_file)
        verification = self.arbiter.verify(project)
        verification_report = formatter.format_verification(verification)
        return verification_report

    def _read_json_object(self, file_name, object_key, ret):
        if not file_name:
            return ret
        else:
            with open(file_name) as fp:
                return json.load(fp)[object_key]

    def _get_arbiter(self):
        licenses_denied_file = self._args.licenses_denied_file
        licenses_preference_file = self._args.licenses_preference_file
        alias_file = self._args.alias_file

        if self._args.licenses_info_file:
            licenses_denied_file = self._args.licenses_info_file
            licenses_preference_file = self._args.licenses_info_file

        licenses_denied = self._read_json_object(licenses_denied_file, "licenses_denied", [])
        licenses_preferences = self._read_json_object(licenses_preference_file, "license_preferences", [])
        arbiter = Arbiter(license_db=self._args.license_matrix_file, licenses_preferences=licenses_preferences, denied_licenses=licenses_denied, alias_file=alias_file)

        return arbiter

    def verify(self):
        formatter = FormatterFactory.formatter(self._args.output_format)

        if self._args.out_license and self._args.in_license_expr:
            compats = self.arbiter.inbounds_outbound_check(self._args.out_license, self._args.in_license_expr)
            formatted = formatter.format_compatibilities(compats)
            return formatted

        elif self._args.verify_flict:
            project_reader = ProjectReaderFactory.get_projectreader(self._args.verify_flict, None, "flict")
            return self._handle_lico_project(project_reader, self._args.verify_flict, formatter)

        elif self._args.verify_sbom:
            project_reader = ProjectReaderFactory.get_projectreader(self._args.verify_sbom, self._args.sbom_dirs)
            return self._handle_lico_project(project_reader, self._args.verify_sbom, formatter)



###################################################################
#
# flict - FOSS License Compatibility Tool
#
# SPDX-FileCopyrightText: 2020, 2022 Henrik Sandklef
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
###################################################################

from flict.flictlib.format.format import FlictFormatter

import json


class JsonFormatter(FlictFormatter):

    def format_support_licenses(self, supported_licenses):
        return json.dumps(supported_licenses)

    def format_license_list(self, license_list):
        return json.dumps(license_list)

    def format_license_combinations(self, project):
        return json.dumps({
            'license_combinations': project.projects_combinations(),
        })

    def format_outbound_license(self, outbound_candidate):
        return json.dumps(outbound_candidate)

    def format_compats(self, compats):
        return json.dumps(compats)

    def format_simplified(self, simplified):
        return json.dumps(simplified)

    def format_verified_license(self, license_expression, outbound_candidate):
        compat = len(outbound_candidate) != 0
        return json.dumps({"license_expression": license_expression,
                           "compatible": compat,
                           "outbound_candidate": outbound_candidate},
                          )

    def format_relicense_information(self, license_handler):
        return json.dumps(license_handler.relicensing_information()['original']['relicense_definitions'])

    def format_compatibilities(self, compats):
        return json.dumps(compats)

    def format_licenses(self, licenses):
        return json.dumps(licenses)

    def format_verification(self, verification):
        return json.dumps(verification)



###################################################################
#
# flict - FOSS License Compatibility Tool
#
# SPDX-FileCopyrightText: 2021 Henrik Sandklef, 2021 Konrad Weihmann
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
###################################################################

from flict.flictlib.format.format import FlictFormatter
from flict.flictlib.format.common import compat_interprets
from flict.flictlib.format.text_format import TextFormatter
from flict.flictlib.license import License

COMPATIBILITY_TAG = "compatibility"

MANIFEST_HEADERS = {
    "projects": "#",
    "packages": "##",
    "package_name": "###",
    "dependencies": "####",
    "problems": "####",
    "dependency": "#####",
    "identified_license": "####",
}


PACKAGE_HEADERS = {
    "projects": "",
    "packages": "#",
    "package_name": "##",
    "dependencies": "###",
    "problems": "###",
    "dependency": "####",
    "identified_license": "###",
}


class MarkdownFlictFormatter(FlictFormatter):

    def __init__(self):
        self.license = License(None)

    def format_compats(self, compats):
        return self.output_compat_markdown(compats)

    def format_relicense_information(self, license_handler):
        return "# Translation information\n" + TextFormatter.format_relicense_information(license_handler)

    def format_translation_information(self, license_handler):
        return "# Translation information\n" + TextFormatter.format_translation_information(license_handler)

    def format_compatibilities(self, compats):
        return compats[COMPATIBILITY_TAG]

    def format_licenses(self, licenses):
        return "\n".join(licenses)

    def format_simplified(self, simplified):
        return f"""
        # Original license\n\n{simplified['original']}
        # Simplified license\n\n{simplified['simplified']}
        """

    def packages_header(self):
        return f"{self.headers['packages']} Packages\n"

    def package_name_header(self, name):
        return f"{self.headers['package_name']} {name}\n"

    def dependencies_header(self):
        return f"{self.headers['dependencies']} Dependencies\n"

    def problem_header(self):
        return f"{self.headers['problems']} Problems\n"

    def dependency_header(self, name):
        return f"{self.headers['dependency']} {name}\n"

    def identified_license_header(self):
        return f"{self.headers['identified_license']} Identified licenses in the combined work\n"

    def output_compat_markdown(self, compats):
        result = []

        result.append("# License compatibilities\n\n")

        result.append("# Licenses\n\n")
        for compat in compats['compatibilities']:
            result.append(" * " + compat['license'] + "\n")

        result.append("\n\n# Compatibilities\n\n")
        for compat in compats['compatibilities']:
            main_license = compat['license']
            result.append(_output_compat_markdown_licenses(main_license, compat))

        return "\n".join(list(set(result)))

    def format_verification(self, verification):
        output = []
        output.append(self.packages_header())
        identified_license = None
        for package in verification['packages']:

            identified_license = package.get('outbound_license', None)
            identified_license_aliased = package.get('outbound_license_aliased', None)

            identified_licenses = package.get('outbound_licenses', None)

            output.append(self.package_name_header(package['name']))
            description = package.get('description', "")
            if description:
                output.append(f"*Description*: {description}\n")

            if identified_licenses == []:
                output.append("Main license: Could not be identified, see problems section below.\n")
                output.append(f"Declared license: {package['license']}\n")
            else:
                # Out of the outbound licenses, present the first and suggest the others
                output.append(f"Main license: {identified_license}\n")

                if identified_license != identified_license_aliased:
                    output.append(f"Main license alias: {identified_license_aliased}\n")
                if identified_licenses and len(identified_licenses) > 1:
                    output.append(f"*Other possible main licenses*: {identified_license}\n".format(identified=", ".join(identified_licenses[1:])))

            output.append(self.dependencies_header())
            for dep in package['dependencies']:

                if not identified_license:
                    dep_identified_license = None
                else:
                    dep_identified_license = self.get_dep_license(dep, identified_license)

                version = dep.get('version', None)

                output.append(self.dependency_header(dep['name']))
                if version:
                    output.append(f"* version: {version}")
                output.append(f"* declared license:  {dep['license']}")
                if dep_identified_license:
                    output.append(f"* concluded license: {dep_identified_license}")
                output.append("\n")

            output.append(self.identified_license_header())
            license_list = list(verification['all_licenses'])
            license_list.sort(key=str.lower)
            for lic in license_list:
                output.append(f"* {lic}\n")

            # Add problem discussion
            if not dep_identified_license:
                output += self._problem("Could not conclude an outbound license for the combined work.\n")

        return "\n".join(output)

    def _problem(self, message):
        return [
            self.problem_header(),
            message,
        ]


class ManifestMarkdownFlictFormatter(MarkdownFlictFormatter):

    def __init__(self):
        self.license = License(None)
        self.headers = MANIFEST_HEADERS


class PackageMarkdownFlictFormatter(MarkdownFlictFormatter):

    def __init__(self):
        self.license = License(None)
        self.headers = PACKAGE_HEADERS


def _compat_to_fmt(comp_left, comp_right, fmt):
    left = compat_interprets['left'][comp_left][fmt]
    right = compat_interprets['right'][comp_right][fmt]
    return str(right) + str(left)


def _compat_to_markdown(left, comp_left, right, comp_right):
    return _compat_to_fmt(comp_left, comp_right, "markdown")


def _output_compat_markdown_licenses(main_license, compat):
    result = []
    for lic in compat['licenses']:
        inner_license = lic['license']
        comp_left = lic['compatible_left']
        comp_right = lic['compatible_right']
        compat_text = _compat_to_markdown(
            main_license, comp_left, inner_license, comp_right)
        result.append(f"{main_license} {compat_text} {inner_license}\n\n")
    return "".join(result)



###################################################################
#
# SPDX-FileCopyrightText: 2021 Henrik Sandklef
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
###################################################################

import logging

logging.basicConfig(level=logging.WARNING)

main_logger = logging.getLogger('flict')

license_logger = logging.getLogger("flict.license")


def setup(debug_license, verbose):
    if debug_license and verbose:
        license_logger.setLevel(logging.DEBUG)
    elif debug_license:
        license_logger.setLevel(logging.INFO)
    else:
        license_logger.setLevel(logging.WARN)

    if verbose:
        main_logger.setLevel(logging.DEBUG)
        logging.basicConfig(level=logging.DEBUG)
    else:
        main_logger.setLevel(logging.WARNING)
        logging.basicConfig(level=logging.WARNING)
